==> xld-sql-tasker-25.1.0-131.113-sources.jar: com/xebialabs/deployit/engine/tasker/repository/sql/SqlActiveTaskRepository.scala <==
package com.xebialabs.deployit.engine.tasker.repository.sql

import ai.digital.configuration.central.deploy.TaskerSystemProperties
import ai.digital.deploy.task.serdes.TaskPekkoSerializer
import ai.digital.deploy.task.status.TaskPathStatusStore
import ai.digital.deploy.task.steplog.TaskStepLogStore
import com.xebialabs.deployit.core.events.{TaskPathStatusDeleteEvent, TaskStepLogDeleteEvent}
import com.xebialabs.deployit.core.sql._
import com.xebialabs.deployit.core.sql.util.queryStringWithInClause
import com.xebialabs.deployit.engine.api.distribution.TaskExecutionWorkerRepository
import com.xebialabs.deployit.engine.api.execution.TaskExecutionState
import com.xebialabs.deployit.engine.tasker.repository.{ActiveTask, ActiveTaskRepository, CrudTaskRepository}
import com.xebialabs.deployit.engine.tasker.{TaskId, TaskSpecification}
import com.xebialabs.deployit.sql.base.schema.{ActiveTaskDataSchema, ActiveTaskMetadataSchema, ActiveTaskSchema, ActiveTaskMetadataKey => metadataKey}
import grizzled.slf4j.Logging
import org.apache.pekko.actor.ActorPath
import org.joda.time.DateTime
import org.springframework.beans.factory.annotation.{Autowired, Qualifier}
import org.springframework.jdbc.core.JdbcTemplate
import org.springframework.stereotype.Component
import org.springframework.transaction.annotation.Transactional

import java.sql.ResultSet
import java.util
import scala.jdk.CollectionConverters._
import scala.util.{Failure, Success, Try}

@Component
@Transactional("mainTransactionManager")
class SqlActiveTaskRepository(@Autowired taskPathStatusStore: TaskPathStatusStore,
                              @Autowired taskStepLogStore: TaskStepLogStore,
                              @Autowired val workerRepository: TaskExecutionWorkerRepository,
                              @Autowired taskerSystemProperties: TaskerSystemProperties,
                              @Autowired @Qualifier("mainJdbcTemplate") override val jdbcTemplate: JdbcTemplate)
                             (@Autowired @Qualifier("mainSchema") override implicit val schemaInfo: SchemaInfo)
  extends SqlCrudTaskRepository(taskPathStatusStore, taskStepLogStore, jdbcTemplate, taskerSystemProperties)
    with ActiveTaskRepository with ActiveTaskQueries with ActiveTaskMetadataQueries with Logging {

  override def store(taskId: TaskId, spec: TaskSpecification, path: ActorPath): Unit = {
    val workerAddress = path.address.toString
    workerRepository.getWorkerByAddress(workerAddress)
      .orElse(throw new IllegalStateException(s"Worker with $workerAddress not found in worker repository"))
      .foreach(worker => {
        spec.getMetadata.put("worker_name", worker.name)
        store(taskId, spec.getDescription, spec.getOwner.getName, worker.id, spec.getMetadata, Some(spec))
      })
  }

  override def changeOwner(taskId: TaskId, newOwner: String): Unit = {
    jdbcTemplate.update(UPDATE_OWNER, newOwner, taskId)
  }

  override def workerAddress(taskId: TaskId): Option[String] = {
    firstOrNone(jdbcTemplate.query(SELECT_WORKER_ADDRESS, (rs: ResultSet, _: Int) => {
      rs.getString(1)
    }, taskId))
  }

  override def ciPathsByDeployments(taskIds: List[TaskId]): Map[String, List[String]] = {
    Option(taskIds) match {
      case Some(taskIds) if taskIds.nonEmpty =>
        schemaInfo.sqlDialect match {
          case DerbyDialect | OracleDialect | Db2Dialect =>
            executeQuery(ORACLE_OR_DERBY_SELECT_CI_PATHS_BY_DEPL_TASK_ID, taskIds)
          case _ =>
            executeQuery(SELECT_CI_PATHS_BY_DEPL_TASK_ID, taskIds)
        }
      case _ => Map.empty
    }
  }

  override def ciPathsByControlTasks(taskIds: List[TaskId]): Map[String, List[String]] = {
    Option(taskIds) match {
      case Some(taskIds) if taskIds.nonEmpty =>
        schemaInfo.sqlDialect match {
          case DerbyDialect | OracleDialect | Db2Dialect =>
            executeQuery(ORACLE_OR_DERBY_SELECT_CI_PATHS_BY_CONTROL_TASK_ID, taskIds)
          case _ =>
            executeQuery(SELECT_CI_PATHS_BY_CONTROL_TASK_ID, taskIds)
        }
      case _ => Map.empty
    }

  }

  override def recordStartDate(taskId: TaskId): Unit = {
    jdbcTemplate.update(SET_START_DATE, toTimestamp(DateTime.now()), taskId)
  }

  override def recordCompletionDate(taskId: TaskId): Unit = {
    jdbcTemplate.update(SET_COMPLETION_DATE, toTimestamp(DateTime.now()), taskId)
  }

  override def updateState(taskId: TaskId, state:String): Unit = {
    jdbcTemplate.update(UPDATE_STATE, state, taskId)
  }

  private def executeQuery(query: String, taskIds: Iterable[TaskId]): Map[String, List[String]] = {
    import ActiveTaskMetadataSchema._
    queryStringWithInClause(taskIds)(query, sqlb"task.$task_id", "taskIdsInClause") {
      jdbcTemplate.query(_, (rs: ResultSet, rowNum: Int) => (rs.getString(1), rs.getString(2)))
        .asScala.toList
    }.groupBy(_._1).view.mapValues(_.map(_._2)).toMap
  }
}

@Transactional("mainTransactionManager")
class SqlCrudTaskRepository(@Autowired val taskPathStatusStore: TaskPathStatusStore,
                            @Autowired val taskStepLogStore: TaskStepLogStore,
                            @Autowired @Qualifier("mainJdbcTemplate") val jdbcTemplate: JdbcTemplate,
                            @Autowired taskerSystemProperties: TaskerSystemProperties)
                           (@Autowired @Qualifier("mainSchema") implicit val schemaInfo: SchemaInfo)
  extends CrudTaskRepository with ActiveTaskQueries with ActiveTaskMetadataQueries with ActiveTaskDataQueries with Logging {

  import ActiveTaskMetadataSchema._
  import ActiveTaskSchema._

  override def store(taskId: TaskId, description: String, owner: String, workerId: Integer, metadata: util.Map[String, String], taskSpecMaybe: Option[TaskSpecification] = None): Unit = {
    jdbcTemplate.update(INSERT, taskId, description, owner, workerId)
    metadata.forEach { (k, v) =>
      jdbcTemplate.update(INSERT_METADATA, taskId, k, v)
    }
    taskSpecMaybe match {
      case Some(taskSpec) if isResilient =>
        jdbcTemplate.update(INSERT_DATA, taskId, taskerSystemProperties.serialization.`type`, specToBinary(taskSpec))
      case _ =>
    }
  }

  override def task(taskId: TaskId, loadFullSpec: Boolean = false): Option[ActiveTask] =
    mapToActiveTasks(jdbcTemplate.queryForList(SELECT_ALL_BY_TASK_ID, taskId).asScala, loadFullSpec).headOption

  override def tasks(loadFullSpec: Boolean = false): List[ActiveTask] =
    mapToActiveTasks(jdbcTemplate.queryForList(SELECT_ALL).asScala, loadFullSpec)

  override def delete(taskId: TaskId): Unit = {
    jdbcTemplate.update(DELETE_DATA, taskId)
    jdbcTemplate.update(DELETE_METADATA, taskId)
    jdbcTemplate.update(DELETE, taskId)
    taskPathStatusStore.sendDeleteEvent(TaskPathStatusDeleteEvent(taskId))
    taskStepLogStore.sendStepLogDeleteEvent(TaskStepLogDeleteEvent(taskId))
  }

  private def mapToActiveTasks(rows: Iterable[util.Map[String, AnyRef]], loadFullSpec: Boolean): List[ActiveTask] = {
    rows
      .groupBy {
        _.get(ActiveTaskSchema.task_id.name).asInstanceOf[String]
      }.map { case (taskId, list) =>
        val firstElement = list.headOption.getOrElse(throw new NoSuchElementException(s"No active task with taskId $taskId"))
        val taskDescription = firstElement.get(description.name).asInstanceOf[String]
        val owner = firstElement.get(task_owner.name).asInstanceOf[String]
        val workerId = asInteger(firstElement.get(worker_id.name))
        val metadata = list.map { m =>
          m.get(metadata_key.name).asInstanceOf[String] ->
            m.get(metadata_value.name).asInstanceOf[String]
        }.toMap.asJava
        if (isResilient) {
          val taskExecutionState = Option(firstElement.get(state.name)).map(state => TaskExecutionState.valueOf(state.asInstanceOf[String])).orNull
          val startDate = Option(firstElement.get(ActiveTaskSchema.startDate.name)).map(mapDateTime).orNull
          val completionDate = Option(firstElement.get(ActiveTaskSchema.completionDate.name)).map(mapDateTime).orNull
          if (loadFullSpec) {
            val specOption = jdbcTemplate.queryForList(SELECT_DATA, taskId).asScala.headOption
            specOption.map { data =>
              import ActiveTaskDataSchema._
              val serializationType = data.get(serialization_type.name).asInstanceOf[String]
              val taskSpecBytes = data.get(specification.name).asInstanceOf[Array[Byte]]
              binaryToSpec(serializationType, taskSpecBytes, taskId)
            } match {
              case Some(taskSpec) => new ActiveTask(taskId, taskDescription, owner, metadata, workerId, taskExecutionState, startDate, completionDate, Some(taskSpec))
              case _ => new ActiveTask(taskId, taskDescription, owner, metadata, workerId, taskExecutionState, startDate, completionDate)
            }
          } else {
            new ActiveTask(taskId, taskDescription, owner, metadata, workerId, taskExecutionState, startDate, completionDate)
          }
        } else {
          new ActiveTask(taskId, taskDescription, owner, metadata, workerId)
        }
      }.toList
  }

  protected lazy val isResilient = taskerSystemProperties.resilient

  protected lazy val serializer: TaskPekkoSerializer =
    TaskPekkoSerializer.getTaskPekkoSerializerByName(taskerSystemProperties.serialization.`type`)

  protected def deserializer(serializationType: String) : TaskPekkoSerializer = {
    TaskPekkoSerializer.getTaskPekkoSerializerByName(serializationType)
  }

  protected def specToBinary(spec: TaskSpecification): Array[Byte] = {
    Try(serializer.toBinary(spec)) match {
      case Success(value) => value
      case Failure(exception) =>
        logger.error(exception.getMessage, exception)
        throw exception
    }
  }

  protected def binaryToSpec(serializationType: String, taskSpecBytes: Array[Byte], taskId: String): TaskSpecification =
    deserializer(serializationType).fromBinary(taskSpecBytes, classOf[TaskSpecification]) match {
      case spec: TaskSpecification => spec
      case _ => throw new RuntimeException(s"Could not deserialize TaskSpecification [$taskId]")
    }
}

trait ActiveTaskQueries extends Queries {
  import ActiveTaskSchema._

  lazy val INSERT = sqlb"insert into $tableName ($task_id, $description, $task_owner, $worker_id) values (?,?,?,?)"
  lazy val UPDATE_OWNER = sqlb"update $tableName set $task_owner = ? where $task_id = ?"
  lazy val SELECT_ALL_BY_TASK_ID: String = {
    import com.xebialabs.deployit.sql.base.schema.{ActiveTaskMetadataSchema => md}
    sqlb"select t.$task_id, t.$description, t.$task_owner, t.$worker_id, t.$state, t.$startDate, t.$completionDate, md.${md.metadata_key}, md.${md.metadata_value} from $tableName t left join ${md.tableName} md on t.$task_id = md.${md.task_id} where t.$task_id = ?"
  }
  lazy val SELECT_WORKER_ADDRESS: String = {
    import com.xebialabs.deployit.engine.tasker.repository.sql.{WorkersSchema => ws}
    sqlb"select ${ws.address} from $tableName t left join ${ws.tableName} w on t.$worker_id = w.${ws.worker_id} where $task_id = ?"
  }
  lazy val SELECT_ALL: String = {
    import com.xebialabs.deployit.sql.base.schema.{ActiveTaskMetadataSchema => md}
    sqlb"select t.$task_id, t.$description, t.$task_owner, t.$worker_id, t.$state, t.$startDate, t.$completionDate, md.${md.metadata_key}, md.${md.metadata_value} from $tableName t left join ${md.tableName} md on t.$task_id = md.${md.task_id}"
  }
  lazy val SET_START_DATE = sqlb"update $tableName set $startDate = ? where $task_id = ?"
  lazy val SET_COMPLETION_DATE = sqlb"update $tableName set $completionDate = ? where $task_id = ?"
  lazy val UPDATE_STATE = sqlb"update $tableName set $state = ? where $task_id = ?"

  lazy val DELETE = sqlb"delete from $tableName where $task_id = ?"
}

trait ActiveTaskMetadataQueries extends Queries {

  import ActiveTaskMetadataSchema._

  lazy val INSERT_METADATA = sqlb"insert into $tableName ($task_id, $metadata_key, $metadata_value) values (?,?,?)"
  lazy val SELECT_METADATA = sqlb"select $metadata_key, $metadata_value from $tableName where $task_id = ?"
  private lazy val SELECT_CI_PATHS_FRAGMENT: String = {
    import com.xebialabs.deployit.sql.base.schema.{CIS => ci}
    sqlb"  select task.$task_id, ci.${ci.path} " +
      sqlb"from $tableName task inner join ${ci.tableName} ci on ci.${ci.ID} = ${schemaInfo.sqlDialect.castToInt(metadata_value.tableAlias("task"))} " +
      sqlb"where (:taskIdsInClause) "
  }
  lazy val SELECT_CI_PATHS_BY_DEPL_TASK_ID: String = {
    sqlb"$SELECT_CI_PATHS_FRAGMENT and (task.$metadata_key = '${metadataKey.ENVIRONMENT_INTERNAL_ID}' or task.$metadata_key = '${metadataKey.APPLICATION_INTERNAL_ID}')"
  }
  lazy val SELECT_CI_PATHS_BY_CONTROL_TASK_ID: String = {
    sqlb" $SELECT_CI_PATHS_FRAGMENT and task.$metadata_key = '${metadataKey.CONTROL_TASK_TARGET_INTERNAL_CI}'"
  }

  private lazy val ORACLE_OR_DERBY_SELECT_CI_PATHS_FRAGMENT: String = {
    import com.xebialabs.deployit.sql.base.schema.{CIS => ci}
    sqlb"  select task.$task_id, ci.${ci.path} " +
      sqlb"from $tableName task inner join ${ci.tableName} ci on cast(ci.${ci.ID} as char(254)) = cast(task.$metadata_value as char(254)) " +
      sqlb"where (:taskIdsInClause) "
  }

  lazy val ORACLE_OR_DERBY_SELECT_CI_PATHS_BY_DEPL_TASK_ID: String = {
    sqlb"$ORACLE_OR_DERBY_SELECT_CI_PATHS_FRAGMENT and (task.$metadata_key = '${metadataKey.ENVIRONMENT_INTERNAL_ID}' or task.$metadata_key = '${metadataKey.APPLICATION_INTERNAL_ID}')"
  }
  lazy val ORACLE_OR_DERBY_SELECT_CI_PATHS_BY_CONTROL_TASK_ID: String = {
    sqlb" $ORACLE_OR_DERBY_SELECT_CI_PATHS_FRAGMENT and task.$metadata_key = '${metadataKey.CONTROL_TASK_TARGET_INTERNAL_CI}'"
  }

  lazy val DELETE_METADATA = sqlb"delete from $tableName where $task_id = ?"
}

trait ActiveTaskDataQueries extends Queries {
  import ActiveTaskDataSchema._

  lazy val INSERT_DATA = sqlb"insert into $tableName ($task_id, $serialization_type, $specification) values (?,?,?)"
  lazy val SELECT_DATA: String = sqlb"select $task_id, $serialization_type, $specification from $tableName where $task_id = ?"
  lazy val DELETE_DATA: String = sqlb"delete from $tableName where $task_id = ?"
}

==> xld-sql-tasker-25.3.3-1215.1110-sources.jar: com/xebialabs/deployit/engine/tasker/repository/sql/SqlActiveTaskRepository.scala <==
package com.xebialabs.deployit.engine.tasker.repository.sql

import ai.digital.configuration.central.deploy.TaskerSystemProperties
import ai.digital.deploy.task.serdes.TaskPekkoSerializer
import ai.digital.deploy.task.status.TaskPathStatusStore
import ai.digital.deploy.task.steplog.TaskStepLogStore
import com.xebialabs.deployit.core.events.{TaskPathStatusDeleteEvent, TaskStepLogDeleteEvent}
import com.xebialabs.deployit.core.sql._
import com.xebialabs.deployit.core.sql.util.queryStringWithInClause
import com.xebialabs.deployit.engine.api.distribution.TaskExecutionWorkerRepository
import com.xebialabs.deployit.engine.api.execution.TaskExecutionState
import com.xebialabs.deployit.engine.tasker.repository.{ActiveTask, ActiveTaskRepository, CrudTaskRepository}
import com.xebialabs.deployit.engine.tasker.{TaskId, TaskSpecification}
import com.xebialabs.deployit.sql.base.schema.{ActiveTaskDataSchema, ActiveTaskMetadataSchema, ActiveTaskSchema, ActiveTaskMetadataKey => metadataKey}
import grizzled.slf4j.Logging
import org.apache.pekko.actor.ActorPath
import org.joda.time.DateTime
import org.springframework.beans.factory.annotation.{Autowired, Qualifier}
import org.springframework.jdbc.core.JdbcTemplate
import org.springframework.stereotype.Component
import org.springframework.transaction.annotation.Transactional

import java.sql.ResultSet
import java.util
import scala.jdk.CollectionConverters._
import scala.util.{Failure, Success, Try}

@Component
@Transactional("mainTransactionManager")
class SqlActiveTaskRepository(@Autowired taskPathStatusStore: TaskPathStatusStore,
                              @Autowired taskStepLogStore: TaskStepLogStore,
                              @Autowired val workerRepository: TaskExecutionWorkerRepository,
                              @Autowired taskerSystemProperties: TaskerSystemProperties,
                              @Autowired @Qualifier("mainJdbcTemplate") override val jdbcTemplate: JdbcTemplate)
                             (@Autowired @Qualifier("mainSchema") override implicit val schemaInfo: SchemaInfo)
  extends SqlCrudTaskRepository(taskPathStatusStore, taskStepLogStore, jdbcTemplate, taskerSystemProperties)
    with ActiveTaskRepository with ActiveTaskQueries with ActiveTaskMetadataQueries with Logging {

  override def store(taskId: TaskId, spec: TaskSpecification, path: ActorPath): Unit = {
    val workerAddress = path.address.toString
    workerRepository.getWorkerByAddress(workerAddress)
      .orElse(throw new IllegalStateException(s"Worker with $workerAddress not found in worker repository"))
      .foreach(worker => {
        spec.getMetadata.put("worker_name", worker.name)
        store(taskId, spec.getDescription, spec.getOwner.getName, worker.id, spec.getMetadata, Some(spec))
      })
  }

  override def changeOwner(taskId: TaskId, newOwner: String): Unit = {
    jdbcTemplate.update(UPDATE_OWNER, newOwner, taskId)
  }

  override def workerAddress(taskId: TaskId): Option[String] = {
    firstOrNone(jdbcTemplate.query(SELECT_WORKER_ADDRESS, (rs: ResultSet, _: Int) => {
      rs.getString(1)
    }, taskId))
  }

  override def ciPathsByDeployments(taskIds: List[TaskId]): Map[String, List[String]] = {
    Option(taskIds) match {
      case Some(taskIds) if taskIds.nonEmpty =>
        schemaInfo.sqlDialect match {
          case DerbyDialect | OracleDialect | Db2Dialect | MssqlDialect =>
            executeQuery(ORACLE_OR_DERBY_SELECT_CI_PATHS_BY_DEPL_TASK_ID, taskIds)
          case _ =>
            executeQuery(SELECT_CI_PATHS_BY_DEPL_TASK_ID, taskIds)
        }
      case _ => Map.empty
    }
  }

  override def ciPathsByControlTasks(taskIds: List[TaskId]): Map[String, List[String]] = {
    Option(taskIds) match {
      case Some(taskIds) if taskIds.nonEmpty =>
        schemaInfo.sqlDialect match {
          case DerbyDialect | OracleDialect | Db2Dialect | MssqlDialect =>
            executeQuery(ORACLE_OR_DERBY_SELECT_CI_PATHS_BY_CONTROL_TASK_ID, taskIds)
          case _ =>
            executeQuery(SELECT_CI_PATHS_BY_CONTROL_TASK_ID, taskIds)
        }
      case _ => Map.empty
    }

  }

  override def recordStartDate(taskId: TaskId): Unit = {
    jdbcTemplate.update(SET_START_DATE, toTimestamp(DateTime.now()), taskId)
  }

  override def recordCompletionDate(taskId: TaskId): Unit = {
    jdbcTemplate.update(SET_COMPLETION_DATE, toTimestamp(DateTime.now()), taskId)
  }

  override def updateState(taskId: TaskId, state:String): Unit = {
    jdbcTemplate.update(UPDATE_STATE, state, taskId)
  }

  private def executeQuery(query: String, taskIds: Iterable[TaskId]): Map[String, List[String]] = {
    import ActiveTaskMetadataSchema._
    queryStringWithInClause(taskIds)(query, sqlb"task.$task_id", "taskIdsInClause") {
      jdbcTemplate.query(_, (rs: ResultSet, rowNum: Int) => (rs.getString(1), rs.getString(2)))
        .asScala.toList
    }.groupBy(_._1).view.mapValues(_.map(_._2)).toMap
  }
}

@Transactional("mainTransactionManager")
class SqlCrudTaskRepository(@Autowired val taskPathStatusStore: TaskPathStatusStore,
                            @Autowired val taskStepLogStore: TaskStepLogStore,
                            @Autowired @Qualifier("mainJdbcTemplate") val jdbcTemplate: JdbcTemplate,
                            @Autowired taskerSystemProperties: TaskerSystemProperties)
                           (@Autowired @Qualifier("mainSchema") implicit val schemaInfo: SchemaInfo)
  extends CrudTaskRepository with ActiveTaskQueries with ActiveTaskMetadataQueries with ActiveTaskDataQueries with Logging {

  import ActiveTaskMetadataSchema._
  import ActiveTaskSchema._

  override def store(taskId: TaskId, description: String, owner: String, workerId: Integer, metadata: util.Map[String, String], taskSpecMaybe: Option[TaskSpecification] = None): Unit = {
    jdbcTemplate.update(INSERT, taskId, description, owner, workerId)
    metadata.forEach { (k, v) =>
      jdbcTemplate.update(INSERT_METADATA, taskId, k, v)
    }
    taskSpecMaybe match {
      case Some(taskSpec) if isResilient =>
        jdbcTemplate.update(INSERT_DATA, taskId, taskerSystemProperties.serialization.`type`, specToBinary(taskSpec))
      case _ =>
    }
  }

  override def task(taskId: TaskId, loadFullSpec: Boolean = false): Option[ActiveTask] =
    mapToActiveTasks(jdbcTemplate.queryForList(SELECT_ALL_BY_TASK_ID, taskId).asScala, loadFullSpec).headOption

  override def tasks(loadFullSpec: Boolean = false): List[ActiveTask] =
    mapToActiveTasks(jdbcTemplate.queryForList(SELECT_ALL).asScala, loadFullSpec)

  override def delete(taskId: TaskId): Unit = {
    jdbcTemplate.update(DELETE_DATA, taskId)
    jdbcTemplate.update(DELETE_METADATA, taskId)
    jdbcTemplate.update(DELETE, taskId)
    taskPathStatusStore.sendDeleteEvent(TaskPathStatusDeleteEvent(taskId))
    taskStepLogStore.sendStepLogDeleteEvent(TaskStepLogDeleteEvent(taskId))
  }

  private def mapToActiveTasks(rows: Iterable[util.Map[String, AnyRef]], loadFullSpec: Boolean): List[ActiveTask] = {
    rows
      .groupBy {
        _.get(ActiveTaskSchema.task_id.name).asInstanceOf[String]
      }.map { case (taskId, list) =>
        val firstElement = list.headOption.getOrElse(throw new NoSuchElementException(s"No active task with taskId $taskId"))
        val taskDescription = firstElement.get(description.name).asInstanceOf[String]
        val owner = firstElement.get(task_owner.name).asInstanceOf[String]
        val workerId = asInteger(firstElement.get(worker_id.name))
        val metadata = list.map { m =>
          m.get(metadata_key.name).asInstanceOf[String] ->
            m.get(metadata_value.name).asInstanceOf[String]
        }.toMap.asJava
        if (isResilient) {
          val taskExecutionState = Option(firstElement.get(state.name)).map(state => TaskExecutionState.valueOf(state.asInstanceOf[String])).orNull
          val startDate = Option(firstElement.get(ActiveTaskSchema.startDate.name)).map(mapDateTime).orNull
          val completionDate = Option(firstElement.get(ActiveTaskSchema.completionDate.name)).map(mapDateTime).orNull
          if (loadFullSpec) {
            val specOption = jdbcTemplate.queryForList(SELECT_DATA, taskId).asScala.headOption
            specOption.map { data =>
              import ActiveTaskDataSchema._
              val serializationType = data.get(serialization_type.name).asInstanceOf[String]
              val taskSpecBytes = data.get(specification.name).asInstanceOf[Array[Byte]]
              binaryToSpec(serializationType, taskSpecBytes, taskId)
            } match {
              case Some(taskSpec) => new ActiveTask(taskId, taskDescription, owner, metadata, workerId, taskExecutionState, startDate, completionDate, Some(taskSpec))
              case _ => new ActiveTask(taskId, taskDescription, owner, metadata, workerId, taskExecutionState, startDate, completionDate)
            }
          } else {
            new ActiveTask(taskId, taskDescription, owner, metadata, workerId, taskExecutionState, startDate, completionDate)
          }
        } else {
          new ActiveTask(taskId, taskDescription, owner, metadata, workerId)
        }
      }.toList
  }

  protected lazy val isResilient = taskerSystemProperties.resilient

  protected lazy val serializer: TaskPekkoSerializer =
    TaskPekkoSerializer.getTaskPekkoSerializerByName(taskerSystemProperties.serialization.`type`)

  protected def deserializer(serializationType: String) : TaskPekkoSerializer = {
    TaskPekkoSerializer.getTaskPekkoSerializerByName(serializationType)
  }

  protected def specToBinary(spec: TaskSpecification): Array[Byte] = {
    Try(serializer.toBinary(spec)) match {
      case Success(value) => value
      case Failure(exception) =>
        logger.error(exception.getMessage, exception)
        throw exception
    }
  }

  protected def binaryToSpec(serializationType: String, taskSpecBytes: Array[Byte], taskId: String): TaskSpecification =
    deserializer(serializationType).fromBinary(taskSpecBytes, classOf[TaskSpecification]) match {
      case spec: TaskSpecification => spec
      case _ => throw new RuntimeException(s"Could not deserialize TaskSpecification [$taskId]")
    }
}

trait ActiveTaskQueries extends Queries {
  import ActiveTaskSchema._

  lazy val INSERT = sqlb"insert into $tableName ($task_id, $description, $task_owner, $worker_id) values (?,?,?,?)"
  lazy val UPDATE_OWNER = sqlb"update $tableName set $task_owner = ? where $task_id = ?"
  lazy val SELECT_ALL_BY_TASK_ID: String = {
    import com.xebialabs.deployit.sql.base.schema.{ActiveTaskMetadataSchema => md}
    sqlb"select t.$task_id, t.$description, t.$task_owner, t.$worker_id, t.$state, t.$startDate, t.$completionDate, md.${md.metadata_key}, md.${md.metadata_value} from $tableName t left join ${md.tableName} md on t.$task_id = md.${md.task_id} where t.$task_id = ?"
  }
  lazy val SELECT_WORKER_ADDRESS: String = {
    import com.xebialabs.deployit.engine.tasker.repository.sql.{WorkersSchema => ws}
    sqlb"select ${ws.address} from $tableName t left join ${ws.tableName} w on t.$worker_id = w.${ws.worker_id} where $task_id = ?"
  }
  lazy val SELECT_ALL: String = {
    import com.xebialabs.deployit.sql.base.schema.{ActiveTaskMetadataSchema => md}
    sqlb"select t.$task_id, t.$description, t.$task_owner, t.$worker_id, t.$state, t.$startDate, t.$completionDate, md.${md.metadata_key}, md.${md.metadata_value} from $tableName t left join ${md.tableName} md on t.$task_id = md.${md.task_id}"
  }
  lazy val SET_START_DATE = sqlb"update $tableName set $startDate = ? where $task_id = ?"
  lazy val SET_COMPLETION_DATE = sqlb"update $tableName set $completionDate = ? where $task_id = ?"
  lazy val UPDATE_STATE = sqlb"update $tableName set $state = ? where $task_id = ?"

  lazy val DELETE = sqlb"delete from $tableName where $task_id = ?"
}

trait ActiveTaskMetadataQueries extends Queries {

  import ActiveTaskMetadataSchema._

  lazy val INSERT_METADATA = sqlb"insert into $tableName ($task_id, $metadata_key, $metadata_value) values (?,?,?)"
  lazy val SELECT_METADATA = sqlb"select $metadata_key, $metadata_value from $tableName where $task_id = ?"
  private lazy val SELECT_CI_PATHS_FRAGMENT: String = {
    import com.xebialabs.deployit.sql.base.schema.{CIS => ci}
    sqlb"  select task.$task_id, ci.${ci.path} " +
      sqlb"from $tableName task inner join ${ci.tableName} ci on ci.${ci.ID} = ${schemaInfo.sqlDialect.castToInt(metadata_value.tableAlias("task"))} " +
      sqlb"where (:taskIdsInClause) "
  }
  lazy val SELECT_CI_PATHS_BY_DEPL_TASK_ID: String = {
    sqlb"$SELECT_CI_PATHS_FRAGMENT and (task.$metadata_key = '${metadataKey.ENVIRONMENT_INTERNAL_ID}' or task.$metadata_key = '${metadataKey.APPLICATION_INTERNAL_ID}')"
  }
  lazy val SELECT_CI_PATHS_BY_CONTROL_TASK_ID: String = {
    sqlb" $SELECT_CI_PATHS_FRAGMENT and task.$metadata_key = '${metadataKey.CONTROL_TASK_TARGET_INTERNAL_CI}'"
  }

  private lazy val ORACLE_OR_DERBY_SELECT_CI_PATHS_FRAGMENT: String = {
    import com.xebialabs.deployit.sql.base.schema.{CIS => ci}
    sqlb"  select task.$task_id, ci.${ci.path} " +
      sqlb"from $tableName task inner join ${ci.tableName} ci on cast(ci.${ci.ID} as char(254)) = cast(task.$metadata_value as char(254)) " +
      sqlb"where (:taskIdsInClause) "
  }

  lazy val ORACLE_OR_DERBY_SELECT_CI_PATHS_BY_DEPL_TASK_ID: String = {
    sqlb"$ORACLE_OR_DERBY_SELECT_CI_PATHS_FRAGMENT and (task.$metadata_key = '${metadataKey.ENVIRONMENT_INTERNAL_ID}' or task.$metadata_key = '${metadataKey.APPLICATION_INTERNAL_ID}')"
  }
  lazy val ORACLE_OR_DERBY_SELECT_CI_PATHS_BY_CONTROL_TASK_ID: String = {
    sqlb" $ORACLE_OR_DERBY_SELECT_CI_PATHS_FRAGMENT and task.$metadata_key = '${metadataKey.CONTROL_TASK_TARGET_INTERNAL_CI}'"
  }

  lazy val DELETE_METADATA = sqlb"delete from $tableName where $task_id = ?"
}

trait ActiveTaskDataQueries extends Queries {
  import ActiveTaskDataSchema._

  lazy val INSERT_DATA = sqlb"insert into $tableName ($task_id, $serialization_type, $specification) values (?,?,?)"
  lazy val SELECT_DATA: String = sqlb"select $task_id, $serialization_type, $specification from $tableName where $task_id = ?"
  lazy val DELETE_DATA: String = sqlb"delete from $tableName where $task_id = ?"
}
